"""Contains RTK models for robot location and RTK positions."""

from dataclasses import dataclass


@dataclass
class Point:
    """Returns a lat long."""

    latitude: float = 0.0
    longitude: float = 0.0

    def __init__(self, latitude: float = 0.0, longitude: float = 0.0) -> None:
        self.latitude = latitude
        self.longitude = longitude


@dataclass
class Dock(Point):
    """Stores robot dock position."""

    rotation: int = 0


@dataclass
class Location:
    """Stores/retrieves RTK GPS data."""

    device: Point
    RTK: Point
    dock: Dock
    position_type: int = 0
    orientation: int = 0  # 360 degree rotation +-
    work_zone: int = 0

    def __init__(self) -> None:
        self.device = Point()
        self.RTK = Point()
        self.dock = Dock()



"""MammotionMQTT."""

import asyncio
import base64
import hashlib
import hmac
import json
import logging
from logging import getLogger
from typing import Awaitable, Callable, Optional

import betterproto
from linkkit.linkkit import LinkKit
from paho.mqtt.client import MQTTMessage

from pymammotion.aliyun.cloud_gateway import CloudIOTGateway
from pymammotion.data.mqtt.event import ThingEventMessage
from pymammotion.data.mqtt.properties import ThingPropertiesMessage
from pymammotion.data.mqtt.status import ThingStatusMessage
from pymammotion.proto.luba_msg import LubaMsg

logger = getLogger(__name__)


class MammotionMQTT:
    """MQTT client for pymammotion."""

    def __init__(
        self,
        region_id: str,
        product_key: str,
        device_name: str,
        device_secret: str,
        iot_token: str,
        cloud_client: CloudIOTGateway,
        client_id: Optional[str] = None,
    ) -> None:
        """Create instance of MammotionMQTT."""
        super().__init__()
        self._cloud_client = cloud_client
        self.is_connected = False
        self.is_ready = False
        self.on_connected: Optional[Callable[[], Awaitable[None]]] = None
        self.on_ready: Optional[Callable[[], Awaitable[None]]] = None
        self.on_error: Optional[Callable[[str], Awaitable[None]]] = None
        self.on_disconnected: Optional[Callable[[], Awaitable[None]]] = None
        self.on_message: Optional[Callable[[str, str, str], Awaitable[None]]] = None

        self._product_key = product_key
        self._device_name = device_name
        self._device_secret = device_secret
        self._iot_token = iot_token
        self._mqtt_username = f"{device_name}&{product_key}"
        # linkkit provides the correct MQTT service for all of this and uses paho under the hood
        if client_id is None:
            client_id = f"python-{device_name}"
        self._mqtt_client_id = f"{client_id}|securemode=2,signmethod=hmacsha1|"
        sign_content = f"clientId{client_id}deviceName{device_name}productKey{product_key}"
        self._mqtt_password = hmac.new(
            device_secret.encode("utf-8"), sign_content.encode("utf-8"), hashlib.sha1
        ).hexdigest()

        self._client_id = client_id
        self.loop = asyncio.get_running_loop()

        self._linkkit_client = LinkKit(
            region_id,
            product_key,
            device_name,
            device_secret,
            auth_type="",
            client_id=client_id,
            password=self._mqtt_password,
            username=self._mqtt_username,
        )

        self._linkkit_client.enable_logger(level=logging.ERROR)
        self._linkkit_client.on_connect = self._thing_on_connect
        self._linkkit_client.on_disconnect = self._on_disconnect
        self._linkkit_client.on_thing_enable = self._thing_on_thing_enable
        self._linkkit_client.on_topic_message = self._thing_on_topic_message
        self._mqtt_host = f"{self._product_key}.iot-as-mqtt.{region_id}.aliyuncs.com"

    def connect_async(self) -> None:
        """Connect async to MQTT Server."""
        logger.info("Connecting...")
        if self._linkkit_client.check_state() is LinkKit.LinkKitState.INITIALIZED:
            self._linkkit_client.thing_setup()
        self._linkkit_client.connect_async()

    def disconnect(self) -> None:
        """Disconnect from MQTT Server."""
        logger.info("Disconnecting...")
        self._linkkit_client.disconnect()

    def _thing_on_thing_enable(self, user_data) -> None:
        """Is called when Thing is enabled."""
        logger.debug("on_thing_enable")
        self.is_connected = True
        # logger.debug('subscribe_topic, topic:%s' % echo_topic)
        # self._linkkit_client.subscribe_topic(echo_topic, 0)
        self._linkkit_client.subscribe_topic(
            f"/sys/{self._product_key}/{self._device_name}/app/down/account/bind_reply"
        )
        self._linkkit_client.subscribe_topic(
            f"/sys/{self._product_key}/{self._device_name}/app/down/thing/event/property/post_reply"
        )
        self._linkkit_client.subscribe_topic(f"/sys/{self._product_key}/{self._device_name}/app/down/thing/events")
        self._linkkit_client.subscribe_topic(f"/sys/{self._product_key}/{self._device_name}/app/down/thing/status")
        self._linkkit_client.subscribe_topic(f"/sys/{self._product_key}/{self._device_name}/app/down/thing/properties")
        self._linkkit_client.subscribe_topic(
            f"/sys/{self._product_key}/{self._device_name}/app/down/thing/model/down_raw"
        )

        self._linkkit_client.publish_topic(
            f"/sys/{self._product_key}/{self._device_name}/app/up/account/bind",
            json.dumps(
                {
                    "id": "msgid1",
                    "version": "1.0",
                    "request": {"clientId": self._mqtt_username},
                    "params": {"iotToken": self._iot_token},
                }
            ),
        )

        if self.on_ready:
            self.is_ready = True
            future = asyncio.run_coroutine_threadsafe(self.on_ready(), self.loop)
            asyncio.wrap_future(future, loop=self.loop)
        # self._linkkit_client.query_ota_firmware()
        # command = MammotionCommand(device_name="Luba")
        # self._cloud_client.send_cloud_command(command.get_report_cfg())

    def _thing_on_topic_message(self, topic, payload, qos, user_data) -> None:
        """Is called when thing topic comes in."""
        logger.debug(
            "on_topic_message, receive message, topic:%s, payload:%s, qos:%d",
            topic,
            payload,
            qos,
        )
        payload = json.loads(payload)
        iot_id = payload.get("params", {}).get("iotId", "")
        if iot_id != "" and self.on_message:
            future = asyncio.run_coroutine_threadsafe(self.on_message(topic, payload, iot_id), self.loop)
            asyncio.wrap_future(future, loop=self.loop)

    def _thing_on_connect(self, session_flag, rc, user_data) -> None:
        """Is called on thing connect."""
        self.is_connected = True
        if self.on_connected is not None:
            future = asyncio.run_coroutine_threadsafe(self.on_connected(), self.loop)
            asyncio.wrap_future(future, loop=self.loop)

        logger.debug("on_connect, session_flag:%d, rc:%d", session_flag, rc)

        # self._linkkit_client.subscribe_topic(f"/sys/{self._product_key}/{self._device_name}/#")

    def _on_disconnect(self, _client, _userdata) -> None:
        """Is called on disconnect."""
        logger.info("Disconnected")
        self.is_connected = False
        self.is_ready = False
        if self.on_disconnected:
            future = asyncio.run_coroutine_threadsafe(self.on_disconnected(), self.loop)
            asyncio.wrap_future(future, loop=self.loop)

    def _on_message(self, _client, _userdata, message: MQTTMessage) -> None:
        """Is called when message is received."""
        logger.info("Message on topic %s", message.topic)

        payload = json.loads(message.payload)
        if message.topic.endswith("/app/down/thing/events"):
            event = ThingEventMessage(**payload)
            params = event.params
            if params.identifier == "device_protobuf_msg_event":
                content = LubaMsg().parse(base64.b64decode(params.value.content.proto))

                logger.info("Unhandled protobuf event: %s", betterproto.which_one_of(content, "LubaSubMsg"))
            elif params.identifier == "device_warning_event":
                logger.debug("identifier event: %s", params.identifier)
            else:
                logger.info("Unhandled event: %s", params.identifier)
        elif message.topic.endswith("/app/down/thing/status"):
            status = ThingStatusMessage(**payload)
            logger.debug(status.params.status.value)
        elif message.topic.endswith("/app/down/thing/properties"):
            properties = ThingPropertiesMessage(**payload)
            logger.debug("properties: %s", properties)
        else:
            logger.debug("Unhandled topic: %s", message.topic)
            logger.debug(payload)

    def get_cloud_client(self) -> CloudIOTGateway:
        """Return internal cloud client."""
        return self._cloud_client



from enum import Enum


class DeviceType(Enum):
    UNKNOWN = (-1, "UNKNOWN", "Unknown")
    RTK = (0, "RTK", "RTK")
    LUBA = (1, "Luba", "Luba 1")
    LUBA_2 = (2, "Luba-VS", "Luba 2")
    LUBA_YUKA = (3, "Yuka-", "Yuka")
    YUKA_MINI = (4, "Yuka-MN", "Yuka Mini")
    YUKA_MINI2 = (5, "Yuka-YM", "Yuka Mini 2")
    LUBA_VP = (6, "Luba-VP", "Luba VP")

    def __init__(self, value: int, name: str, model: str) -> None:
        self._value = value
        self._name = name
        self._model = model

    def get_name(self) -> str:
        return self._name

    def get_model(self):
        return self._model

    def get_value(self):
        return self._value

    def get_value_str(self):
        return str(self._value)

    def set_value(self, value) -> None:
        self._value = value

    @staticmethod
    def valueof(value):
        """Return the corresponding DeviceType based on the input value.

        This function takes an integer value as input and returns the
        corresponding DeviceType enum value.

        Args:
            value (int): An integer representing the device type.

        Returns:
            DeviceType: The corresponding DeviceType enum value based on the input value.

        """

        if value == 0:
            return DeviceType.RTK
        elif value == 1:
            return DeviceType.LUBA
        elif value == 2:
            return DeviceType.LUBA_2
        elif value == 3:
            return DeviceType.LUBA_YUKA
        elif value == 4:
            return DeviceType.YUKA_MINI
        elif value == 5:
            return DeviceType.YUKA_MINI2
        else:
            return DeviceType.UNKNOWN

    @staticmethod
    def value_of_str(device_name: str, product_key: str = ""):
        """Determine the type of device based on the provided device name and
        product key.

        Args:
            device_name (str): The name of the device.
            product_key (str?): The product key associated with the device. Defaults to "".

        Returns:
            DeviceType: The type of device based on the provided information.

        """

        if not device_name and not product_key:
            return DeviceType.UNKNOWN

        try:
            substring = device_name[:3]
            substring2 = device_name[:7]

            if DeviceType.RTK.get_name() in substring or DeviceType.contain_rtk_product_key(product_key):
                return DeviceType.RTK
            elif DeviceType.LUBA_2.get_name() in substring2 or DeviceType.contain_luba_2_product_key(product_key):
                return DeviceType.LUBA_2
            elif DeviceType.LUBA_YUKA.get_name() in substring2:
                return DeviceType.LUBA_YUKA
            elif DeviceType.YUKA_MINI.get_name() in substring2:
                return DeviceType.YUKA_MINI
            elif DeviceType.YUKA_MINI2.get_name() in substring2:
                return DeviceType.YUKA_MINI2
            elif DeviceType.LUBA.get_name() in substring2 or DeviceType.contain_luba_product_key(product_key):
                return DeviceType.LUBA
            else:
                print("unknown device type")
                return DeviceType.UNKNOWN
        except Exception:
            return DeviceType.UNKNOWN

    @staticmethod
    def has_4g(device_name: str, product_key: str = ""):
        """Check if the device has 4G capability based on the device name and
        optional product key.

        This function determines the device type based on the device name and
        product key (if provided). It then checks if the device type has a value
        greater than or equal to the 4G threshold.

        Args:
            device_name (str): The name of the device.
            product_key (str?): The product key associated with the device. Defaults to "".

        Returns:
            bool: True if the device has 4G capability, False otherwise.

        """

        if not product_key:
            device_type = DeviceType.value_of_str(device_name)
        else:
            device_type = DeviceType.value_of_str(device_name, product_key)

        return device_type.get_value() >= DeviceType.LUBA_2.get_value()

    @staticmethod
    def is_luba1(device_name: str, product_key: str = ""):
        """Check if the given device is of type LUBA.

        This function determines if the device specified by 'device_name' is of
        type LUBA. If 'product_key' is provided, it is used to further identify
        the device type.

        Args:
            device_name (str): The name of the device.
            product_key (str?): The product key associated with the device. Defaults to "".

        Returns:
            bool: True if the device is of type LUBA, False otherwise.

        """

        if not product_key:
            device_type = DeviceType.value_of_str(device_name)
        else:
            device_type = DeviceType.value_of_str(device_name, product_key)

        return device_type.get_value() == DeviceType.LUBA.get_value()

    @staticmethod
    def is_luba_2(device_name: str, product_key: str = ""):
        """Check if the device type is LUBA 2 or higher based on the device name
        and optional product key.

        Args:
            device_name (str): The name of the device.
            product_key (str?): The product key associated with the device. Defaults to "".

        Returns:
            bool: True if the device type is LUBA 2 or higher, False otherwise.

        """

        if not product_key:
            device_type = DeviceType.value_of_str(device_name)
        else:
            device_type = DeviceType.value_of_str(device_name, product_key)

        return device_type.get_value() >= DeviceType.LUBA_2.get_value()

    @staticmethod
    def is_yuka(device_name: str):
        """Check if the given device name corresponds to a LUBA_YUKA device type.

        Args:
            device_name (str): The name of the device to be checked.

        Returns:
            bool: True if the device type is LUBA_YUKA, False otherwise.

        """

        return (
            DeviceType.value_of_str(device_name).get_value() == DeviceType.LUBA_YUKA.get_value()
            or DeviceType.value_of_str(device_name).get_value() == DeviceType.YUKA_MINI.get_value()
            or DeviceType.value_of_str(device_name).get_value() == DeviceType.YUKA_MINI2.get_value()
        )

    @staticmethod
    def is_rtk(device_name: str, product_key: str = ""):
        """Check if the device type is within the range of RTK devices.

        This function determines if the device type corresponding to the given
        device name and optional product key falls within the range of RTK
        (Real-Time Kinematic) devices.

        Args:
            device_name (str): The name of the device.
            product_key (str?): The product key associated with the device. Defaults to "".

        Returns:
            bool: True if the device type is within the RTK range, False otherwise.

        """

        if not product_key:
            device_type = DeviceType.value_of_str(device_name)
        else:
            device_type = DeviceType.value_of_str(device_name, product_key)

        return DeviceType.RTK.get_value() <= device_type.get_value() < DeviceType.LUBA.get_value()

    @staticmethod
    def contain_rtk_product_key(product_key) -> bool:
        """Check if the given product key is in a predefined list of RTK product
        keys.

        Args:
            product_key (str): The product key to be checked.

        Returns:
            bool: True if the product key is in the predefined list, False otherwise.

        """

        if not product_key:
            return False
        return product_key in ["a1qXkZ5P39W", "a1Nc68bGZzX"]

    @staticmethod
    def contain_luba_product_key(product_key) -> bool:
        """Check if the given product key is in the list of valid product keys.

        Args:
            product_key (str): The product key to be checked.

        Returns:
            bool: True if the product key is in the list of valid keys, False otherwise.

        """

        if not product_key:
            return False
        return product_key in [
            "a1UBFdq6nNz",
            "a1x0zHD3Xop",
            "a1pvCnb3PPu",
            "a1kweSOPylG",
            "a1JFpmAV5Ur",
            "a1BmXWlsdbA",
            "a1jOhAYOIG8",
            "a1K4Ki2L5rK",
            "a1ae1QnXZGf",
            "a1nf9kRBWoH",
            "a1ZU6bdGjaM",
        ]

    @staticmethod
    def contain_luba_2_product_key(product_key) -> bool:
        """Check if the given product key is present in a predefined list.

        Args:
            product_key (str): The product key to be checked.

        Returns:
            bool: True if the product key is in the predefined list, False otherwise.

        """

        if not product_key:
            return False
        return product_key in ["a1iMygIwxFC", "a1LLmy1zc0j", "a1LLmy1zc0j"]

    @staticmethod
    def contain_yuka_product_key(product_key) -> bool:
        """Check if the given product key is present in a predefined list.

        Args:
            product_key (str): The product key to be checked.

        Returns:
            bool: True if the product key is in the predefined list, False otherwise.

        """

        if not product_key:
            return False
        return product_key in ["a1IQV0BrnXb"]

    def is_support_video(self):
        return self != DeviceType.LUBA



from dataclasses import asdict, dataclass, field


@dataclass
class ConnectData:
    connect_type: int = 0
    ble_rssi: int = 0
    wifi_rssi: int = 0

    @classmethod
    def from_dict(cls, data: dict) -> "ConnectData":
        return cls(
            connect_type=data.get("connect_type", 0),
            ble_rssi=data.get("ble_rssi", 0),
            wifi_rssi=data.get("wifi_rssi", 0),
        )


@dataclass
class DeviceData:
    sys_status: int = 0
    charge_state: int = 0
    battery_val: int = 0
    sensor_status: int = 0
    last_status: int = 0
    sys_time_stamp: str = ""

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            sys_status=data.get("sys_status", 0),
            charge_state=data.get("charge_state", 0),
            battery_val=data.get("battery_val", 0),
            sensor_status=data.get("sensor_status", 0),
            last_status=data.get("last_status", 0),
            sys_time_stamp=data.get("sys_time_stamp", ""),
        )


@dataclass
class RTKData:
    status: int = 0
    pos_level: int = 0
    gps_stars: int = 0
    dis_status: str = ""
    co_view_stars: int = 0

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            status=data.get("status", 0),
            pos_level=data.get("pos_level", 0),
            gps_stars=data.get("gps_stars", 0),
            dis_status=data.get("dis_status", ""),
            co_view_stars=data.get("co_view_stars", 0),
        )


@dataclass
class LocationData:
    real_pos_x: int = 0
    real_pos_y: int = 0
    real_toward: int = 0
    pos_type: int = 0
    bol_hash: str = ""

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            real_pos_x=data.get("real_pos_x", 0),
            real_pos_y=data.get("real_pos_y", 0),
            real_toward=data.get("real_toward", 0),
            pos_type=data.get("pos_type", 0),
            bol_hash=data.get("bol_hash", ""),
        )


@dataclass
class WorkData:
    path: int = 0
    path_hash: str = ""
    progress: int = 0
    area: int = 0
    bp_info: int = 0
    bp_hash: str = ""
    bp_pos_x: int = 0
    bp_pos_y: int = 0
    real_path_num: str = ""
    path_pos_x: int = 0
    path_pos_y: int = 0
    ub_zone_hash: str = ""
    ub_path_hash: str = ""
    init_cfg_hash: str = ""
    ub_ecode_hash: str = ""
    nav_run_mode: int = 0
    test_mode_status: int = 0
    man_run_speed: int = 0
    nav_edit_status: int = 0
    knife_height: int = 0

    @classmethod
    def from_dict(cls, data: dict):
        return cls(
            path=data.get("path", 0),
            path_hash=data.get("path_hash", ""),
            progress=data.get("progress", 0),
            area=data.get("area", 0),
            bp_info=data.get("bp_info", 0),
            bp_hash=data.get("bp_hash", ""),
            bp_pos_x=data.get("bp_pos_x", 0),
            bp_pos_y=data.get("bp_pos_y", 0),
            real_path_num=data.get("real_path_num", ""),
            path_pos_x=data.get("path_pos_x", 0),
            path_pos_y=data.get("path_pos_y", 0),
            ub_zone_hash=data.get("ub_zone_hash", ""),
            ub_path_hash=data.get("ub_path_hash", ""),
            init_cfg_hash=data.get("init_cfg_hash", ""),
            ub_ecode_hash=data.get("ub_ecode_hash", ""),
            nav_run_mode=data.get("nav_run_mode", 0),
            test_mode_status=data.get("test_mode_status", 0),
            man_run_speed=data.get("man_run_speed", 0),
            nav_edit_status=data.get("nav_edit_status", 0),
            knife_height=data.get("knife_height", 0),
        )


@dataclass
class ReportData:
    connect: ConnectData = field(default_factory=ConnectData)
    dev: DeviceData = field(default_factory=DeviceData)
    rtk: RTKData = field(default_factory=RTKData)
    locations: list[LocationData] = field(default_factory=list)
    work: WorkData = field(default_factory=WorkData)

    def from_dict(self, data: dict):
        locations = self.locations
        if data.get("locations") is not None:
            locations = [LocationData.from_dict(loc) for loc in data.get("locations", [])]

        return ReportData(
            connect=ConnectData.from_dict(data.get("connect", asdict(self.connect))),
            dev=DeviceData.from_dict(data.get("dev", asdict(self.dev))),
            rtk=RTKData.from_dict(data.get("rtk", asdict(self.rtk))),
            locations=locations,
            work=WorkData.from_dict(data.get("work", asdict(self.work))),
        )



from dataclasses import dataclass
from typing import Generic, Literal, Optional, TypeVar

from mashumaro import DataClassDictMixin
from mashumaro.mixins.orjson import DataClassORJSONMixin

DataT = TypeVar("DataT")


@dataclass
class ErrorInfo:
    code: str
    platform: str
    module: str
    variant: str
    level: str
    description: str
    en_implication: str
    en_solution: str
    zh_implication: str
    zh_solution: str
    de_implication: str
    de_solution: str
    fr_implication: str
    fr_solution: str
    it_implication: str
    it_solution: str
    es_implication: str
    es_solution: str
    cs_implication: str
    cs_solution: str
    sk_implication: str
    sk_solution: str
    pl_implication: str
    pl_solution: str
    nl_implication: str
    nl_solution: str
    da_implication: str
    da_solution: str
    sv_implication: str
    sv_solution: str
    sl_implication: str
    sl_solution: str
    pt_implication: str
    pt_solution: str


@dataclass
class Response(DataClassDictMixin, Generic[DataT]):
    code: int
    msg: str
    data: DataT | None = None


@dataclass
class LoginResponseUserInformation(DataClassORJSONMixin):
    areaCode: str
    domainAbbreviation: str
    email: Optional[str]
    userId: str
    userAccount: str
    authType: str


@dataclass
class LoginResponseData(DataClassORJSONMixin):
    access_token: str
    token_type: Literal["bearer"]
    refresh_token: str
    expires_in: int
    scope: Literal["read"]
    grant_type: Literal["password"]
    authorization_code: str
    userInformation: LoginResponseUserInformation
    jti: str



"""Device control of mammotion robots over bluetooth or MQTT."""

from __future__ import annotations

import asyncio
import logging
from enum import Enum
from functools import cache
from typing import Any

from aiohttp import ClientSession
from bleak.backends.device import BLEDevice

from pymammotion.aliyun.cloud_gateway import CloudIOTGateway
from pymammotion.aliyun.dataclass.dev_by_account_response import Device
from pymammotion.const import MAMMOTION_DOMAIN
from pymammotion.data.model.account import Credentials
from pymammotion.data.model.device import MowingDevice
from pymammotion.http.http import connect_http
from pymammotion.mammotion.devices.mammotion_bluetooth import MammotionBaseBLEDevice
from pymammotion.mammotion.devices.mammotion_cloud import MammotionBaseCloudDevice, MammotionCloud
from pymammotion.mqtt import MammotionMQTT

TIMEOUT_CLOUD_RESPONSE = 10

_LOGGER = logging.getLogger(__name__)


class ConnectionPreference(Enum):
    """Enum for connection preference."""

    EITHER = 0
    WIFI = 1
    BLUETOOTH = 2


class MammotionMixedDeviceManager:
    _ble_device: MammotionBaseBLEDevice | None = None
    _cloud_device: MammotionBaseCloudDevice | None = None
    _mowing_state: MowingDevice = MowingDevice()

    def __init__(
        self,
        name: str,
        cloud_device: Device | None = None,
        ble_device: BLEDevice | None = None,
        mqtt: MammotionCloud | None = None,
    ) -> None:
        self.name = name
        self.add_ble(ble_device)
        self.add_cloud(cloud_device, mqtt)

    def mower_state(self):
        return self._mowing_state

    def ble(self) -> MammotionBaseBLEDevice | None:
        return self._ble_device

    def cloud(self) -> MammotionBaseCloudDevice | None:
        return self._cloud_device

    def add_ble(self, ble_device: BLEDevice) -> None:
        if ble_device is not None:
            self._ble_device = MammotionBaseBLEDevice(self._mowing_state, ble_device)

    def add_cloud(self, cloud_device: Device | None = None, mqtt: MammotionCloud | None = None) -> None:
        if cloud_device is not None:
            self._cloud_device = MammotionBaseCloudDevice(
                mqtt, cloud_device=cloud_device, mowing_state=self._mowing_state
            )

    def replace_cloud(self, cloud_device: MammotionBaseCloudDevice) -> None:
        self._cloud_device = cloud_device

    def replace_ble(self, ble_device: MammotionBaseBLEDevice) -> None:
        self._ble_device = ble_device

    def has_cloud(self) -> bool:
        return self._cloud_device is not None

    def has_ble(self) -> bool:
        return self._ble_device is not None


class MammotionDevices:
    devices: dict[str, MammotionMixedDeviceManager] = {}

    def add_device(self, mammotion_device: MammotionMixedDeviceManager) -> None:
        exists: MammotionMixedDeviceManager | None = self.devices.get(mammotion_device.name)
        if exists is None:
            self.devices[mammotion_device.name] = mammotion_device
            return
        if mammotion_device.has_cloud():
            exists.replace_cloud(mammotion_device.cloud())
        if mammotion_device.has_ble():
            exists.replace_ble(mammotion_device.ble())

    def get_device(self, mammotion_device_name: str) -> MammotionMixedDeviceManager:
        return self.devices.get(mammotion_device_name)


async def create_devices(
    ble_device: BLEDevice,
    cloud_credentials: Credentials | None = None,
    preference: ConnectionPreference = ConnectionPreference.BLUETOOTH,
):
    mammotion = Mammotion(ble_device, preference)

    if cloud_credentials and preference == ConnectionPreference.EITHER or preference == ConnectionPreference.WIFI:
        cloud_client = await Mammotion.login(
            cloud_credentials.account_id or cloud_credentials.email, cloud_credentials.password
        )
        await mammotion.initiate_cloud_connection(cloud_client)

    return mammotion


@cache
class Mammotion:
    """Represents a Mammotion account and its devices."""

    devices = MammotionDevices()
    cloud_client: CloudIOTGateway | None = None
    mqtt: MammotionCloud | None = None

    def __init__(
        self, ble_device: BLEDevice, preference: ConnectionPreference = ConnectionPreference.BLUETOOTH
    ) -> None:
        """Initialize MammotionDevice."""
        if ble_device:
            self.devices.add_device(MammotionMixedDeviceManager(name=ble_device.name, ble_device=ble_device))

        if preference:
            self._preference = preference

    async def initiate_cloud_connection(self, cloud_client: CloudIOTGateway) -> None:
        if self.mqtt is not None:
            if self.mqtt.is_connected:
                return

        self.cloud_client = cloud_client
        self.mqtt = MammotionCloud(
            MammotionMQTT(
                region_id=cloud_client.region_response.data.regionId,
                product_key=cloud_client.aep_response.data.productKey,
                device_name=cloud_client.aep_response.data.deviceName,
                device_secret=cloud_client.aep_response.data.deviceSecret,
                iot_token=cloud_client.session_by_authcode_response.data.iotToken,
                client_id=cloud_client.client_id,
                cloud_client=cloud_client,
            )
        )

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.mqtt.connect_async)

        for device in cloud_client.devices_by_account_response.data.data:
            if device.deviceName.startswith(("Luba-", "Yuka-")) and self.devices.get_device(device.deviceName) is None:
                self.devices.add_device(
                    MammotionMixedDeviceManager(name=device.deviceName, cloud_device=device, mqtt=self.mqtt)
                )

    def set_disconnect_strategy(self, disconnect: bool) -> None:
        for device_name, device in self.devices.devices:
            if device.ble() is not None:
                ble_device: MammotionBaseBLEDevice = device.ble()
                ble_device.set_disconnect_strategy(disconnect)

    @staticmethod
    async def login(account: str, password: str) -> CloudIOTGateway:
        """Login to mammotion cloud."""
        cloud_client = CloudIOTGateway()
        async with ClientSession(MAMMOTION_DOMAIN) as session:
            mammotion_http = await connect_http(account, password)
            country_code = mammotion_http.login_info.userInformation.domainAbbreviation
            _LOGGER.debug("CountryCode: " + country_code)
            _LOGGER.debug("AuthCode: " + mammotion_http.login_info.authorization_code)
            loop = asyncio.get_running_loop()
            cloud_client.set_http(mammotion_http)
            await loop.run_in_executor(
                None, cloud_client.get_region, country_code, mammotion_http.login_info.authorization_code
            )
            await cloud_client.connect()
            await cloud_client.login_by_oauth(country_code, mammotion_http.login_info.authorization_code)
            await loop.run_in_executor(None, cloud_client.aep_handle)
            await loop.run_in_executor(None, cloud_client.session_by_auth_code)

            await loop.run_in_executor(None, cloud_client.list_binding_by_account)
            return cloud_client

    def get_device_by_name(self, name: str) -> MammotionMixedDeviceManager:
        return self.devices.get_device(name)

    async def send_command(self, name: str, key: str):
        """Send a command to the device."""
        device = self.get_device_by_name(name)
        if device:
            if self._preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().command(key)
            if self._preference is ConnectionPreference.WIFI:
                return await device.cloud().command(key)
            # TODO work with both with EITHER

    async def send_command_with_args(self, name: str, key: str, **kwargs: Any):
        """Send a command with args to the device."""
        device = self.get_device_by_name(name)
        if device:
            if self._preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().command(key, **kwargs)
            if self._preference is ConnectionPreference.WIFI:
                return await device.cloud().command(key, **kwargs)
            # TODO work with both with EITHER

    async def start_sync(self, name: str, retry: int):
        device = self.get_device_by_name(name)
        if device:
            if self._preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().start_sync(retry)
            if self._preference is ConnectionPreference.WIFI:
                return await device.cloud().start_sync(retry)
            # TODO work with both with EITHER

    async def start_map_sync(self, name: str):
        device = self.get_device_by_name(name)
        if device:
            if self._preference is ConnectionPreference.BLUETOOTH:
                return await device.ble().start_map_sync()
            if self._preference is ConnectionPreference.WIFI:
                return await device.cloud().start_map_sync()
            # TODO work with both with EITHER

    def mower(self, name: str):
        device = self.get_device_by_name(name)
        if device:
            return device.mower_state()



from typing import List, Optional


class RegionData:
    def __init__(self) -> None:
        self.hash: Optional[int] = None
        self.action: int = 0
        self.current_frame: int = 0
        self.data_hash: Optional[int] = None
        self.data_len: int = 0
        self.p_hash_a: Optional[int] = None
        self.p_hash_b: Optional[int] = None
        self.path: Optional[list[list[float]]] = None
        self.pver: int = 0
        self.result: int = 0
        self.sub_cmd: int = 0
        self.total_frame: int = 0
        self.type: int = 0

    def set_hash(self, hash: int) -> None:
        self.hash = hash

    def get_data_len(self) -> int:
        return self.data_len

    def set_data_len(self, data_len: int) -> None:
        self.data_len = data_len

    def get_pver(self) -> int:
        return self.pver

    def set_pver(self, pver: int) -> None:
        self.pver = pver

    def get_sub_cmd(self) -> int:
        return self.sub_cmd

    def set_sub_cmd(self, sub_cmd: int) -> None:
        self.sub_cmd = sub_cmd

    def get_result(self) -> int:
        return self.result

    def set_result(self, result: int) -> None:
        self.result = result

    def get_action(self) -> int:
        return self.action

    def set_action(self, action: int) -> None:
        self.action = action

    def get_type(self) -> int:
        return self.type

    def set_type(self, type: int) -> None:
        self.type = type

    def get_total_frame(self) -> int:
        return self.total_frame

    def set_total_frame(self, total_frame: int) -> None:
        self.total_frame = total_frame

    def get_current_frame(self) -> int:
        return self.current_frame

    def set_current_frame(self, current_frame: int) -> None:
        self.current_frame = current_frame

    def get_path(self) -> Optional[list[list[float]]]:
        return self.path

    def set_path(self, path: list[list[float]]) -> None:
        self.path = path

    def get_hash(self) -> Optional[int]:
        return self.hash

    def set_data_hash(self, data_hash: int) -> None:
        self.data_hash = data_hash

    def get_data_hash(self) -> Optional[int]:
        return self.data_hash

    def set_p_hash_a(self, p_hash_a: int) -> None:
        self.p_hash_a = p_hash_a

    def get_p_hash_a(self) -> Optional[int]:
        return self.p_hash_a

    def set_p_hash_b(self, p_hash_b: int) -> None:
        self.p_hash_b = p_hash_b

    def get_p_hash_b(self) -> Optional[int]:
        return self.p_hash_b

    def __str__(self) -> str:
        return f"RegionData{{pver={self.pver}, sub_cmd={self.sub_cmd}, result={self.result}, action={self.action}, type={self.type}, Hash={self.hash}, total_frame={self.total_frame}, current_frame={self.current_frame}, data_hash={self.data_hash}, p_hash_a={self.p_hash_a}, p_hash_b={self.p_hash_b}, data_len={self.data_len}, path={self.path}}}"



import asyncio


class Event:
    def __init__(self) -> None:
        self.__eventhandlers = []

    def __iadd__(self, handler):
        self.__eventhandlers.append(handler)
        return self

    def __isub__(self, handler):
        self.__eventhandlers.remove(handler)
        return self

    async def __call__(self, *args, **kwargs) -> None:
        await asyncio.gather(*[handler(*args, **kwargs) for handler in self.__eventhandlers])


class MoveEvent:
    def __init__(self) -> None:
        self.OnMoveFinished = Event()

    async def MoveFinished(self) -> None:
        # This function will be executed once blufi finishes after a movement command and will
        # raise an event
        await self.OnMoveFinished()

    def AddSubscribersForMoveFinishedEvent(self, objMethod) -> None:
        self.OnMoveFinished += objMethod

    def RemoveSubscribersForMoveFinishedEvent(self, objMethod) -> None:
        self.OnMoveFinished -= objMethod


class BleNotificationEvent:
    def __init__(self) -> None:
        self.OnBleNotification = Event()

    async def BleNotification(self, data: bytearray) -> None:
        # This function will be executed when data is received.
        await self.OnBleNotification(data)

    def AddSubscribersForBleNotificationEvent(self, objMethod) -> None:
        self.OnBleNotification += objMethod

    def RemoveSubscribersForBleNotificationEvent(self, objMethod) -> None:
        self.OnBleNotification -= objMethod


class DataEvent:
    """Callbacks for data events."""

    def __init__(self) -> None:
        self.on_data_event = Event()

    async def data_event(self, data) -> None:
        # This function will be executed when data is received.
        await self.on_data_event(data)

    def add_subscribers(self, obj_method) -> None:
        self.on_data_event += obj_method

    def remove_subscribers(self, obj_method) -> None:
        self.on_data_event -= obj_method



# === sendOrderMsg_Net  ===

from pymammotion.mammotion.commands.messages.navigation import MessageNavigation
from pymammotion.proto import dev_net_pb2, luba_msg_pb2


class MessageNetwork:
    messageNavigation: MessageNavigation = MessageNavigation()

    @staticmethod
    def send_order_msg_net(build) -> bytes:
        luba_msg = luba_msg_pb2.LubaMsg(
            msgtype=luba_msg_pb2.MSG_CMD_TYPE_ESP,
            sender=luba_msg_pb2.DEV_MOBILEAPP,
            rcver=luba_msg_pb2.DEV_COMM_ESP,
            msgattr=luba_msg_pb2.MSG_ATTR_REQ,
            seqs=1,
            version=1,
            subtype=1,
            net=build,
        )

        return luba_msg.SerializeToString()

    def send_todev_ble_sync(self, sync_type: int) -> bytes:
        comm_esp = dev_net_pb2.DevNet(todev_ble_sync=sync_type)
        return self.send_order_msg_net(comm_esp)

    def get_device_base_info(self):
        net = dev_net_pb2.DevNet(todev_devinfo_req=dev_net_pb2.DrvDevInfoReq())
        net.todev_devinfo_req.req_ids.add(id=1, type=6)

        return self.send_order_msg_net(net)

    def get_4g_module_info(self) -> bytes:
        build = dev_net_pb2.DevNet(todev_get_mnet_cfg_req=dev_net_pb2.DevNet().todev_get_mnet_cfg_req)
        print("Send command -- Get device 4G network module information")
        return self.send_order_msg_net(build)

    def get_4g_info(self) -> bytes:
        build = dev_net_pb2.DevNet(todev_mnet_info_req=dev_net_pb2.DevNet().todev_mnet_info_req)
        print("Send command -- Get device 4G network information")
        return self.send_order_msg_net(build)

    def set_zmq_enable(self) -> bytes:
        build = dev_net_pb2.DevNet(
            todev_set_dds2zmq=dev_net_pb2.DrvDebugDdsZmq(
                is_enable=True,
                rx_topic_name="perception_post_result",
                tx_zmq_url="tcp://0.0.0.0:5555",
            )
        )
        print("Send command -- Set vision ZMQ to enable")
        return self.send_order_msg_net(build)

    def set_iot_setting(self, iot_control_type: dev_net_pb2.iot_conctrl_type) -> bytes:
        build = dev_net_pb2.DevNet(todev_set_iot_offline_req=iot_control_type)
        print("Send command -- Device re-online")
        return self.send_order_msg_net(build)

    def set_device_log_upload(
        self,
        request_id: str,
        operation: int,
        server_ip: int,
        server_port: int,
        number: int,
        type: int,
    ) -> bytes:
        build = dev_net_pb2.DrvUploadFileToAppReq(
            bizId=request_id,
            operation=operation,
            serverIp=server_ip,
            serverPort=server_port,
            num=number,
            type=type,
        )
        print(
            f"Send log====Feedback====Command======requestID:{request_id} operation:{operation} serverIp:{server_ip} type:{type}"
        )
        return self.send_order_msg_net(dev_net_pb2.DevNet(todev_ble_sync=1, todev_uploadfile_req=build))

    def set_device_socket_request(
        self,
        request_id: str,
        operation: int,
        server_ip: int,
        server_port: int,
        number: int,
        type: int,
    ) -> bytes:
        """Set device socket request (bluetooth only)."""
        build = dev_net_pb2.DrvUploadFileToAppReq(
            bizId=request_id,
            operation=operation,
            serverIp=server_ip,
            serverPort=server_port,
            num=number,
            type=type,
        )
        print(
            f"Send log====Feedback====Command======requestID:{request_id}  operation:{operation} serverIp:{server_ip}  type:{type}"
        )
        return self.send_order_msg_net(dev_net_pb2.DevNet(todev_ble_sync=1, todev_uploadfile_req=build))

    def get_device_log_info(self, biz_id: str, type: int, log_url: str) -> bytes:
        """Get device log info (bluetooth only)."""
        return self.send_order_msg_net(
            dev_net_pb2.DevNet(
                todev_ble_sync=1,
                todev_req_log_info=dev_net_pb2.DrvUploadFileReq(
                    bizId=biz_id,
                    type=type,
                    url=log_url,
                    num=0,
                    userId="",  # TODO supply user id
                ),
            )
        )

    def cancel_log_update(self, biz_id: str) -> bytes:
        """Cancel log update (bluetooth only)."""
        return self.send_order_msg_net(
            dev_net_pb2.DevNet(todev_log_data_cancel=dev_net_pb2.DrvUploadFileCancel(bizId=biz_id))
        )

    def get_device_network_info(self) -> bytes:
        build = dev_net_pb2.DevNet(todev_networkinfo_req=dev_net_pb2.GetNetworkInfoReq(req_ids=1))
        print("Send command - get device network information")
        return self.send_order_msg_net(build)

    def set_device_4g_enable_status(self, new_4g_status: bool) -> bytes:
        build = dev_net_pb2.DevNet(
            todev_ble_sync=1,
            todev_set_mnet_cfg_req=dev_net_pb2.SetMnetCfgReq(
                cfg=dev_net_pb2.MnetCfg(
                    type=dev_net_pb2.NET_TYPE_WIFI,
                    inet_enable=new_4g_status,
                    mnet_enable=new_4g_status,
                )
            ),
        )

        print(f"Send command - set 4G (on/off status). newWifiStatus={new_4g_status}")
        return self.send_order_msg_net(build)

    def set_device_wifi_enable_status(self, new_wifi_status: bool) -> bytes:
        build = dev_net_pb2.DevNet(
            todev_ble_sync=1,
            todev_Wifi_Configuration=dev_net_pb2.DrvWifiSet(configParam=4, wifi_enable=new_wifi_status),
        )
        print(f"szNetwork: Send command - set network (on/off status). newWifiStatus={new_wifi_status}")
        return self.send_order_msg_net(build)

    def wifi_connectinfo_update(self, device_name: str, is_binary: bool) -> bytes:
        print(
            f"Send command - get Wifi connection information.wifiConnectinfoUpdate().deviceName={device_name}.isBinary={is_binary}"
        )
        if is_binary:
            build = dev_net_pb2.DevNet(
                todev_ble_sync=1,
                todev_WifiMsgUpload=dev_net_pb2.DrvWifiUpload(wifi_msg_upload=1),
            )
            print("Send command - get Wifi connection information")
            return self.send_order_msg_net(build)
        self.wifi_connectinfo_update2()

    def wifi_connectinfo_update2(self) -> None:
        hash_map = {"getMsgCmd": 1}
        # self.post_custom_data(self.get_json_string(
        #     68, hash_map))  # ToDo: Fix this

    def get_record_wifi_list(self, is_binary: bool) -> bytes:
        print(f"getRecordWifiList().isBinary={is_binary}")
        if is_binary:
            build = dev_net_pb2.DevNet(todev_ble_sync=1, todev_WifiListUpload=dev_net_pb2.DrvWifiList())
            print("Send command - get memorized WiFi list upload command")
            return self.send_order_msg_net(build)
        self.get_record_wifi_list2()

    def get_record_wifi_list2(self) -> None:
        pass
        # self.messageNavigation.post_custom_data(
        #     self.get_json_string(69))  # ToDo: Fix this

    def close_clear_connect_current_wifi(self, ssid: str, status: int, is_binary: bool) -> bytes:
        if is_binary:
            build = dev_net_pb2.DevNet(
                todev_ble_sync=1,
                todev_Wifi_Configuration=dev_net_pb2.DrvWifiSet(configParam=status, Confssid=ssid),
            )
            print(
                f"Send command - set network (disconnect, direct connect, forget, no operation reconnect) operation command (downlink ssid={ssid}, status={status})"
            )
            return self.send_order_msg_net(build)
        self.close_clear_connect_current_wifi2(ssid, status)

    def close_clear_connect_current_wifi2(self, ssid: str, get_msg_cmd: int) -> None:
        data = {"ssid": ssid, "getMsgCmd": get_msg_cmd}
        # self.messageNavigation.post_custom_data(
        # ToDo: Fix this
        # self.get_json_string(bleOrderCmd.close_clear_connect_current_wifi, data).encode())



"""MowingDevice class to wrap around the betterproto dataclasses."""

from dataclasses import dataclass

import betterproto

from pymammotion.data.model import HashList, RapidState
from pymammotion.data.model.device_config import DeviceLimits
from pymammotion.data.model.location import Location
from pymammotion.data.model.report_info import ReportData
from pymammotion.proto.dev_net import DevNet
from pymammotion.proto.luba_msg import LubaMsg
from pymammotion.proto.luba_mul import SocMul
from pymammotion.proto.mctrl_driver import MctlDriver
from pymammotion.proto.mctrl_nav import MctlNav
from pymammotion.proto.mctrl_ota import MctlOta
from pymammotion.proto.mctrl_pept import MctlPept
from pymammotion.proto.mctrl_sys import (
    MctlSys,
    MowToAppInfoT,
    ReportInfoData,
    SystemRapidStateTunnelMsg,
    SystemUpdateBufMsg,
)
from pymammotion.utility.constant import WorkMode
from pymammotion.utility.conversions import parse_double
from pymammotion.utility.map import CoordinateConverter


@dataclass
class MowingDevice:
    """Wraps the betterproto dataclasses, so we can bypass the groups for keeping all data."""

    device: LubaMsg
    map: HashList
    location: Location
    mowing_state: RapidState

    def __init__(self) -> None:
        self.device = LubaMsg()
        self.map = HashList(area={}, path={}, obstacle={}, hashlist=[])
        self.location = Location()
        self.report_data = ReportData()
        self.err_code_list = []
        self.err_code_list_time = []
        self.limits = DeviceLimits(30, 70, 0.2, 0.6)
        self.mowing_state = RapidState()

    @classmethod
    def from_raw(cls, raw: dict) -> "MowingDevice":
        """Take in raw data to hold in the betterproto dataclass."""
        mowing_device = MowingDevice()
        mowing_device.device = LubaMsg(**raw)
        return mowing_device

    def update_raw(self, raw: dict) -> None:
        """Update the raw LubaMsg data."""
        self.device = LubaMsg(**raw)

    def buffer(self, buffer_list: SystemUpdateBufMsg) -> None:
        """Update the device based on which buffer we are reading from."""
        match buffer_list.update_buf_data[0]:
            case 1:
                # 4 speed
                self.location.RTK.latitude = parse_double(buffer_list.update_buf_data[5], 8.0)
                self.location.RTK.longitude = parse_double(buffer_list.update_buf_data[6], 8.0)
                self.location.dock.latitude = parse_double(buffer_list.update_buf_data[7], 4.0)
                self.location.dock.longitude = parse_double(buffer_list.update_buf_data[8], 4.0)
                self.location.dock.rotation = buffer_list.update_buf_data[3] + 180
            case 2:
                self.err_code_list.clear()
                self.err_code_list_time.clear()
                self.err_code_list.extend(
                    [
                        buffer_list.update_buf_data[3],
                        buffer_list.update_buf_data[5],
                        buffer_list.update_buf_data[7],
                        buffer_list.update_buf_data[9],
                        buffer_list.update_buf_data[11],
                        buffer_list.update_buf_data[13],
                        buffer_list.update_buf_data[15],
                        buffer_list.update_buf_data[17],
                        buffer_list.update_buf_data[19],
                        buffer_list.update_buf_data[21],
                    ]
                )
                self.err_code_list_time.extend(
                    [
                        buffer_list.update_buf_data[4],
                        buffer_list.update_buf_data[6],
                        buffer_list.update_buf_data[8],
                        buffer_list.update_buf_data[10],
                        buffer_list.update_buf_data[12],
                        buffer_list.update_buf_data[14],
                        buffer_list.update_buf_data[16],
                        buffer_list.update_buf_data[18],
                        buffer_list.update_buf_data[20],
                        buffer_list.update_buf_data[22],
                    ]
                )

    def update_report_data(self, toapp_report_data: ReportInfoData) -> None:
        coordinate_converter = CoordinateConverter(self.location.RTK.latitude, self.location.RTK.longitude)
        for index, location in enumerate(toapp_report_data.locations):
            if index == 0 and location.real_pos_y != 0:
                self.location.position_type = location.pos_type
                self.location.orientation = location.real_toward / 10000
                self.location.device = coordinate_converter.enu_to_lla(
                    parse_double(location.real_pos_y, 4.0), parse_double(location.real_pos_x, 4.0)
                )
                if location.zone_hash:
                    self.location.work_zone = (
                        location.zone_hash if self.report_data.dev.sys_status == WorkMode.MODE_WORKING else 0
                    )

        self.report_data = self.report_data.from_dict(toapp_report_data.to_dict(casing=betterproto.Casing.SNAKE))

    def run_state_update(self, rapid_state: SystemRapidStateTunnelMsg) -> None:
        self.mowing_state = RapidState().from_raw(rapid_state.rapid_state_data)

    def mow_info(self, toapp_mow_info: MowToAppInfoT) -> None:
        pass

    def report_missing_data(self) -> None:
        """Report missing data so we can refetch it."""

    @property
    def net(self):
        """Will return a wrapped betterproto of net."""
        return DevNetData(net=self.device.net)

    @property
    def sys(self):
        """Will return a wrapped betterproto of sys."""
        return SysData(sys=self.device.sys)

    @property
    def nav(self):
        """Will return a wrapped betterproto of nav."""
        return NavData(nav=self.device.nav)

    @property
    def driver(self):
        """Will return a wrapped betterproto of driver."""
        return DriverData(driver=self.device.driver)

    @property
    def mul(self):
        """Will return a wrapped betterproto of mul."""
        return MulData(mul=self.device.mul)

    @property
    def ota(self):
        """Will return a wrapped betterproto of ota."""
        return OtaData(ota=self.device.ota)

    @property
    def pept(self):
        """Will return a wrapped betterproto of pept."""
        return PeptData(pept=self.device.pept)


@dataclass
class DevNetData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    net: dict

    def __init__(self, net: DevNet) -> None:
        if isinstance(net, dict):
            self.net = net
        else:
            self.net = net.to_dict()

    def __getattr__(self, item):
        """Intercept call to get net in dict and return a betterproto dataclass."""
        if self.net.get(item) is None:
            return DevNet().__getattribute__(item)

        if not isinstance(self.net.get(item), dict):
            return self.net.get(item)

        return DevNet().__getattribute__(item).from_dict(value=self.net.get(item))


@dataclass
class SysData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    sys: dict

    def __init__(self, sys: MctlSys) -> None:
        if isinstance(sys, dict):
            self.sys = sys
        else:
            self.sys = sys.to_dict()

    def __getattr__(self, item: str):
        """Intercept call to get sys in dict and return a betterproto dataclass."""
        if self.sys.get(item) is None:
            return MctlSys().__getattribute__(item)

        if not isinstance(self.sys.get(item), dict):
            return self.sys.get(item)

        return MctlSys().__getattribute__(item).from_dict(value=self.sys.get(item))


@dataclass
class NavData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    nav: dict

    def __init__(self, nav: MctlNav) -> None:
        if isinstance(nav, dict):
            self.nav = nav
        else:
            self.nav = nav.to_dict()

    def __getattr__(self, item: str):
        """Intercept call to get nav in dict and return a betterproto dataclass."""
        if self.nav.get(item) is None:
            return MctlNav().__getattribute__(item)

        if not isinstance(self.nav.get(item), dict):
            return self.nav.get(item)

        return MctlNav().__getattribute__(item).from_dict(value=self.nav.get(item))


@dataclass
class DriverData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    driver: dict

    def __init__(self, driver: MctlDriver) -> None:
        if isinstance(driver, dict):
            self.driver = driver
        else:
            self.driver = driver.to_dict()

    def __getattr__(self, item: str):
        """Intercept call to get driver in dict and return a betterproto dataclass."""
        if self.driver.get(item) is None:
            return MctlDriver().__getattribute__(item)

        if not isinstance(self.driver.get(item), dict):
            return self.driver.get(item)

        return MctlDriver().__getattribute__(item).from_dict(value=self.driver.get(item))


@dataclass
class MulData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    mul: dict

    def __init__(self, mul: SocMul) -> None:
        if isinstance(mul, dict):
            self.mul = mul
        else:
            self.mul = mul.to_dict()

    def __getattr__(self, item: str):
        """Intercept call to get mul in dict and return a betterproto dataclass."""
        if self.mul.get(item) is None:
            return SocMul().__getattribute__(item)

        if not isinstance(self.mul.get(item), dict):
            return self.mul.get(item)

        return SocMul().__getattribute__(item).from_dict(value=self.mul.get(item))


@dataclass
class OtaData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    ota: dict

    def __init__(self, ota: MctlOta) -> None:
        if isinstance(ota, dict):
            self.ota = ota
        else:
            self.ota = ota.to_dict()

    def __getattr__(self, item: str):
        """Intercept call to get ota in dict and return a betterproto dataclass."""
        if self.ota.get(item) is None:
            return MctlOta().__getattribute__(item)

        if not isinstance(self.ota.get(item), dict):
            return self.ota.get(item)

        return MctlOta().__getattribute__(item).from_dict(value=self.ota.get(item))


@dataclass
class PeptData:
    """Wrapping class around LubaMsg to return a dataclass from the raw dict."""

    pept: dict

    def __init__(self, pept: MctlPept) -> None:
        if isinstance(pept, dict):
            self.pept = pept
        else:
            self.pept = pept.to_dict()

    def __getattr__(self, item: str):
        """Intercept call to get pept in dict and return a betterproto dataclass."""
        if self.pept.get(item) is None:
            return MctlPept().__getattribute__(item)

        if not isinstance(self.pept.get(item), dict):
            return self.pept.get(item)

        return MctlPept().__getattribute__(item).from_dict(value=self.pept.get(item))



from dataclasses import dataclass, field
from datetime import time

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class SessionOauthToken(DataClassORJSONMixin):
    identityId: str
    refreshTokenExpire: int
    iotToken: str
    iotTokenExpire: int
    refreshToken: str


@dataclass
class SessionByAuthCodeResponse(DataClassORJSONMixin):
    code: int
    data: SessionOauthToken
